/*
 * SMPDataFlowAnalysis.h - <see below>.
 *
 * Copyright (c) 2000, 2001, 2010 - University of Virginia 
 *
 * This file is part of the Memory Error Detection System (MEDS) infrastructure.
 * This file may be used and modified for non-commercial purposes as long as 
 * all copyright, permission, and nonwarranty notices are preserved.  
 * Redistribution is prohibited without prior written consent from the University 
 * of Virginia.
 *
 * Please contact the authors for restrictions applying to commercial use.
 *
 * THIS SOURCE IS PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF
 * MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 *
 * Author: University of Virginia
 * e-mail: jwd@virginia.com
 * URL   : http://www.cs.virginia.edu/
 *
 * Additional copyrights 2010, 2011, 2012, 2013, 2014, 2015 by Zephyr Software LLC
 * e-mail: {clc,jwd}@zephyr-software.com
 * URL   : http://www.zephyr-software.com/
 *
 */

#ifndef SMPDATAFLOWANALYSIS_H
#define SMPDATAFLOWANALYSIS_H 1

// SMPDataFlowAnalysis.h
//
// This header defines the common enumerations and types needed by
//  instructions, basic blocks, and functions.

#include <list>
#include <vector>
#include <map>
#include <set>
#include <string>
#include <utility>

#include <cstddef>
#include <cstdint>

#include "interfaces/SMPDBInterface.h"
#include "interfaces/abstract/all.h"

class SMPFunction;
class SMPBasicBlock;
class SMPInstr;
class DefOrUseList;
class SMPPhiFunction;
class STARSBitSet;


// IDA_SDK_VERSION is defined in pro.h starting in version 5.2 or 5.7 of the SDK.
#ifndef IDA_SDK_VERSION
#define IDA_SDK_VERSION 660  // To make Visual Studio display code properly
#endif

// Insert a floating no-op instruction at top of each function to hold SSA DEFs
//  of LiveIn names?
#define SMP_USE_SSA_FNOP_MARKER 1

// Should we try to propagate memory location types until we encounter
//  an indirect memory access or other potential alias in the DEF-USE chain?
#define SMP_PROPAGATE_MEM_TYPES 1

// Instruction disassembly text.
#define MAXSMPDISASMSTR 1024
// Global var names
#define MAXSMPGLOBALVARSTR 64
// Local var names
#define MAXSMPVARSTR 32
// String for dead registers in SMPInstr
#define MAXSMPDEADREGSSTR 48

// Value for an SSA subscript number before it is initialized by SSA renaming.
#define SMP_SSA_UNINIT  (-1)

// Bit masks for extracting bits from a STARSBitSet unsigned char.
extern const unsigned char STARSBitMasks[8];

// Map register number to a string for printing or annotations.
extern const char *RegNames[];
const char *MDGetRegName(const STARSOpndTypePtr &RegOp); // Distinguishes subword regs from their parent regs
const char *MDGetRegNumName(uint16_t RegNum, uint16_t ByteWidth); // Distinguishes subword regs from their parent regs
extern const unsigned char RegSizes[]; // size in bytes of each reg
extern const char RegDtyps[]; // IDA Pro dtyp (register width) field for each reg number
unsigned char GetRegSize(uint16_t RegNum);

// Define a machine-dependent constant for the EFLAGS register in the x86 architecture.
//  IDA Pro distinguishes among four different flag bits; we aggregate them for most purposes,
//  but separate them for SCCP (Sparse Conditional Constant Propagation).
#define X86_FLAGS_REG STARS_x86_R_efl
#define MD_FLAGS_REG X86_FLAGS_REG

// Define various important registers for machine-dependent use.
#define X86_CARRY_FLAG STARS_x86_R_cf
#define MD_CARRY_FLAG X86_CARRY_FLAG

#define X86_ZERO_FLAG STARS_x86_R_zf
#define MD_ZERO_FLAG X86_ZERO_FLAG

#define X86_SIGN_FLAG STARS_x86_R_sf
#define MD_SIGN_FLAG X86_SIGN_FLAG

#define X86_OVERFLOW_FLAG STARS_x86_R_of
#define MD_OVERFLOW_FLAG X86_OVERFLOW_FLAG

#define X86_INSTRUCTION_POINTER_REG STARS_x86_R_ip
#define MD_INSTRUCTION_POINTER_REG X86_INSTRUCTION_POINTER_REG

#define X86_STACK_POINTER_REG STARS_x86_R_sp
#define MD_STACK_POINTER_REG X86_STACK_POINTER_REG

#define X86_FRAME_POINTER_REG STARS_x86_R_bp
#define MD_FRAME_POINTER_REG X86_FRAME_POINTER_REG

#define X86_RETURN_VALUE_REG STARS_x86_R_ax
#define MD_RETURN_VALUE_REG X86_RETURN_VALUE_REG

#define FIRST_X86_SUBWORD_REG STARS_x86_R_al
#define LAST_X86_SUBWORD_REG STARS_x86_R_dil

#define X86_FIRST_FP_STACK_REG STARS_x86_R_st0
#define MD_FIRST_FP_STACK_REG X86_FIRST_FP_STACK_REG

#define X86_LAST_REG_NO STARS_x86_R_last
#define MD_LAST_REG_NO X86_LAST_REG_NO

#define X86_CALL_INSTRUCTION STARS_NN_call
#define MD_CALL_INSTRUCTION X86_CALL_INSTRUCTION

#define X86_MOVE_INSTRUCTION STARS_NN_mov
#define MD_MOVE_INSTRUCTION X86_MOVE_INSTRUCTION

#define X86_DEFAULT_RETURN_ADDRESS_SIZE global_STARS_program->GetSTARS_ISA_Bytewidth()
#define MD_DEFAULT_RETURN_ADDRESS_SIZE X86_DEFAULT_RETURN_ADDRESS_SIZE

#define X86_BINARY_NOP_INSTRUCTION 0x90
#define MD_BINARY_NOP_INSTRUCTION X86_BINARY_NOP_INSTRUCTION

// Clone operands that will be changed (e.g. normalized) later. Subword regs will be normalized to full-word regs
//  in DEF and USE and dataflow analysis sets, and stack accesses will be normalized in these same sets to be relative
//  to the incoming stack pointer value. These normalizations are needed for proper SSA form, e.g. we need to know that
//  changes to a subword reg also change the full-word reg, and [stack_ptr+0] and [stack_ptr+8] could be the same stack
//  location in a procedure if stack_ptr changes value. The deep-copied clone can be changed without affecting the
//  original operand. If we are only canonicalizing subword regs to prepare for a USE or DEF search, we use
//  CloneIfSubwordReg(). If we are placing operands in dataflow sets and would also later need to normalize stack accesses
//  in addition to normalizing subword regs, we call CloneIfNecessary().

// If CurrOp would change when reg is canonicalized, then return CurrOp->clone(), else return CurrOp;
STARSOpndTypePtr CloneIfSubwordReg(const STARSOpndTypePtr &CurrOp);

// If CurrOp would change when Canonicalized or stack-normalized, then return CurrOp->clone(), else return CurrOp;
STARSOpndTypePtr CloneIfNecessary(const STARSOpndTypePtr &CurrOp, bool UseFP);

// Debug: print one operand from an instruction or DEF or USE list.
void PrintDefUse(unsigned long feature, int OpNum);
void PrintSIB(const STARSOpndTypePtr &Opnd);
void AnnotPrintSIB(const STARSOpndTypePtr &Opnd, bool HasOffset, FILE *OutFile);
void SPARKAnnotPrintSIB(const STARSOpndTypePtr &Opnd, bool HasOffset, FILE *OutFile, uint16_t SegReg, bool UseFP);
void PrintOneOperand(const STARSOpndTypePtr &Opnd, uint32_t features, int OpNum); 
void PrintListOperand(const STARSOpndTypePtr &Opnd, int SSANum = SMP_SSA_UNINIT); 
void PrintOperand(const STARSOpndTypePtr &Opnd);

// Annotations: concisely print one operand.
void AnnotPrintOperand(const STARSOpndTypePtr &Opnd, FILE *OutFile);

// Print opcode string.
void PrintOpcode(uint16_t opcode, FILE *OutFile);

// MACHINE DEPENDENT: Could operand be an indirect memory access?
bool MDIsIndirectMemoryOpnd(const STARSOpndTypePtr &CurrOp, bool UseFP);

// MACHINE DEPENDENT: Extract base reg from SIB byte of operand.
int MD_STARS_sib_base(const STARSOpndTypePtr &x);

// MACHINE DEPENDENT: Extract index reg from SIB byte of operand.
short MD_STARS_sib_index(const STARSOpndTypePtr &x);

// Is CurrOp a memory operand?
bool IsMemOperand(const STARSOpndTypePtr &CurrOp);

// MACHINE DEPENDENT: Is CurrOp the flags register?
bool MDIsFlagsReg(const STARSOpndTypePtr &CurrOp);

// MACHINE DEPENDENT: Is operand a stack memory access?
bool MDIsStackAccessOpnd(const STARSOpndTypePtr &CurrOp, bool UseFP);

// MACHINE DEPENDENT: Is operand a direct stack memory access?
bool MDIsDirectStackAccessOpnd(const STARSOpndTypePtr &CurrOp, bool UseFP);

// MACHINE DEPENDENT: Is operand trackable in data flow analyses (i.e. a stack memory access or a register?)
bool MDIsDataFlowOpnd(const STARSOpndTypePtr &CurrOp, bool UseFP);

// MACHINE DEPENDENT: Is register a stack pointer or frame pointer?
bool MDIsStackPtrReg(int RegNumber, bool UseFP);

// MACHINE DEPENDENT: Is operand a caller-saved register?
bool MDIsCallerSavedReg(const STARSOpndTypePtr &CurrOp);

// MACHINE DEPENDENT: Get base and index registers and displacement/addr from operand.
void MDExtractAddressFields(const STARSOpndTypePtr &MemOp, int &BaseReg, int &IndexReg, uint16_t &Scale, STARS_ea_t &Offset);

// MACHINE DEPENDENT: Is operand type a known type that we want to analyze?
bool MDKnownOperandType(const STARSOpndTypePtr &TempOp);

// MACHINE DEPENDENT: Convert subword register to its enclosing register.
uint16_t MDCanonicalizeSubReg(const uint16_t Reg1);

// If TempOp is a register, call MDCanonicalizeSubReg() on it.
void CanonicalizeOpnd(STARSOpndTypePtr &TempOp);

// Is RegOp the stack pointer or frame pointer register
bool MDIsStackOrFramePointerReg(const STARSOpndTypePtr &RegOp, bool UseFP);

// MACHINE DEPENDENT: Ordering function for register enum values, to use in set containers.
bool MDLessReg(const uint16_t Reg1, const uint16_t Reg2);

// MACHINE DEPENDENT: Ordering function for register enum values, to use in set containers.
bool MDEqReg(const uint16_t Reg1, const uint16_t Reg2);

// MACHINE DEPENDENT: Ordering function for register operands, to use in set containers.
bool MDLessRegOpnd(const STARSOpndTypePtr &RegOp1, const STARSOpndTypePtr &RegOp2);

// Is CurrOp a general purpose register? (not flags, instruction pointer, non-integer reg, etc.)
bool MDIsGeneralPurposeReg(const STARSOpndTypePtr &CurrOp);

// Hash a global name and SSA number into an int, for use in SMPFunction.GlobalDefAddrBySSA map
int HashGlobalNameAndSSA(const STARSOpndTypePtr &DefOp, int SSANum);

// Are operands equal?
bool IsEqOp(const STARSOpndTypePtr &Opnd1, const STARSOpndTypePtr &Opnd2);

// Are operands equal, ignoring bitwidth differences for register operands?
bool IsEqOpIgnoreBitwidth(const STARSOpndTypePtr &Opnd1, const STARSOpndTypePtr &Opnd2);

// Comparison class to use in sorted containers of IDA Pro addresses (type ea_t).
class LessAddr {
public:
	bool operator()(const STARS_ea_t &Addr1, const STARS_ea_t &Addr2) const {
		return (Addr1 < Addr2);
	}
};

// Comparison class for sorted containers of register operands to search for.
//  First item in each pair is a code address where we are about to search, and the
//  second item in each pair is a hash of SSANum and register number obtained from
//  the HashGlobalNameAndSSA() function.
class LessSearchOperand {
public:
	bool operator()(const std::pair<STARS_ea_t, int> &SearchPair1, const std::pair<STARS_ea_t, int> &SearchPair2) const {
		return ((SearchPair1.first < SearchPair2.first)
			|| ((SearchPair1.first == SearchPair2.first) && (SearchPair1.second < SearchPair2.second)));
	}
};

// MACHINE DEPENDENT: comparison class to permit sorting of STARSOpndTypePtr operands.
class LessOp {
public:
	bool operator()(const STARSOpndTypePtr &Opnd1, const STARSOpndTypePtr &Opnd2) const {
#if 1
		return ((*Opnd1) < (*Opnd2));  // use < operator from STARS_op_t
#else
		if (Opnd1.type != Opnd2.type)
			return (Opnd1.type < Opnd2.type);
		switch (Opnd1.type) {
			case o_void: return false;
			case o_reg: return MDLessReg(Opnd1.reg, Opnd2.reg);
			case o_mem: return (Opnd1.addr < Opnd2.addr);
			case o_phrase: if (Opnd1.hasSIB && Opnd2.hasSIB) return ((Opnd1.sib < Opnd2.sib) || ((Opnd1.sib == Opnd2.sib) && (Opnd1.specflag4 < Opnd2.specflag4)));
							else if (Opnd2.hasSIB) return true;  // no SIB < has SIB
							else if (Opnd1.hasSIB) return false; // no SIB < has SIB
							else return MDLessReg(Opnd1.phrase, Opnd2.phrase); // no SIB bytes
			case o_displ: if (Opnd1.hasSIB && Opnd2.hasSIB)
							  return ((Opnd1.sib < Opnd2.sib) 
								|| ((Opnd1.sib == Opnd2.sib) 
									&& ((Opnd1.addr < Opnd2.addr) || ((Opnd1.addr == Opnd2.addr) && (Opnd1.specflag4 < Opnd2.specflag4)))));
							else if (Opnd2.hasSIB) return true;  // no SIB < has SIB
							else if (Opnd1.hasSIB) return false; // no SIB < has SIB
							else return ((Opnd1.addr < Opnd2.addr)
								|| ((Opnd1.addr == Opnd2.addr) 
								&& MDLessReg(Opnd1.reg, Opnd2.reg))); // no SIB bytes
			case o_imm: return (Opnd1.value < Opnd2.value);
			case o_far:  // fall through to o_near case
			case o_near: return (Opnd1.addr < Opnd2.addr);
			case o_trreg:  // fall through
			case o_dbreg:  // fall through
			case o_crreg:  // fall through
			case o_fpreg:  // fall through
			case o_mmxreg: // fall through
			case o_xmmreg: 
			case o_ymmreg: return (Opnd1.reg < Opnd2.reg); // no subword regs to deal with

			default: msg("ERROR: Unknown operand type in LessOp.\n"); return false;
		}; // end switch (Opnd1.type)
#endif
	} // end operator
}; // end class LessOp

// STL ordered container less-than operator for shared_ptr to STARS_op_t objects.
class STARSLessOp {
public:
	bool operator() (const STARSOpndTypePtr &Ptr1, const STARSOpndTypePtr &Ptr2) const {
		return ((*Ptr1) < (*Ptr2)); // use < operator from STARS_op_t
	}
}; // end class STARSLessOp

typedef std::set<STARSOpndTypePtr, STARSLessOp> STARSOpndSet;
typedef STARSOpndSet::iterator STARSOpndSetIter;

// MACHINE DEPENDENT: comparison class to permit sorting of definitions, each of which is a std::pair<STARSOpndTypePtr, STARS_ea_t>.
class LessDefinition {
public:
	bool operator()(const std::pair<STARSOpndTypePtr, STARS_ea_t> &Defn1, const std::pair<STARSOpndTypePtr, STARS_ea_t> &Defn2) const {
		STARSOpndTypePtr Opnd1 = Defn1.first;
		STARSOpndTypePtr Opnd2 = Defn2.first;
		if (IsEqOp(Opnd1, Opnd2)) {
			// If definitions are of the same operand, sort by instruction address of the definition.
			return (Defn1.second < Defn2.second);
		}
		else if (Opnd1->IsRegOp() && Opnd2->IsRegOp()) {
			return MDLessRegOpnd(Opnd1, Opnd2);
		}
#if 1
		return ((*Opnd1) < (*Opnd2));  // use < operator from STARS_op_t
#else
		if (Opnd1.type != Opnd2.type)
			return (Opnd1.type < Opnd2.type);
		switch (Opnd1.type) {
			case o_void: return false;
			case o_reg: return MDLessRegOpnd(Opnd1, Opnd2);
			case o_mem: return (Opnd1.addr < Opnd2.addr);
			case o_phrase: if (Opnd1.hasSIB && Opnd2.hasSIB) return ((Opnd1.sib < Opnd2.sib) || ((Opnd1.sib == Opnd2.sib) && (Opnd1.specflag4 < Opnd2.specflag4)));
							else if (Opnd2.hasSIB) return true;  // no SIB < has SIB
							else if (Opnd1.hasSIB) return false; // no SIB < has SIB
							else return MDLessReg(Opnd1.phrase, Opnd2.phrase); // no SIB bytes
			case o_displ: if (Opnd1.hasSIB && Opnd2.hasSIB)
							  return ((Opnd1.sib < Opnd2.sib) 
								|| ((Opnd1.sib == Opnd2.sib) 
									&& ((Opnd1.addr < Opnd2.addr) || ((Opnd1.addr == Opnd2.addr) && (Opnd1.specflag4 < Opnd2.specflag4)))));
							else if (Opnd2.hasSIB) return true;  // no SIB < has SIB
							else if (Opnd1.hasSIB) return false; // no SIB < has SIB
							else return ((Opnd1.addr < Opnd2.addr)
								|| ((Opnd1.addr == Opnd2.addr) 
								&& MDLessReg(Opnd1.reg, Opnd2.reg))); // no SIB bytes
			case o_imm: return (Opnd1.value < Opnd2.value);
			case o_far:  // fall through to o_near case
			case o_near: return (Opnd1.addr < Opnd2.addr);
			case o_trreg:  // fall through
			case o_dbreg:  // fall through
			case o_crreg:  // fall through
			case o_fpreg:  // fall through
			case o_mmxreg: // fall through
			case o_xmmreg: 
			case o_ymmreg: return (Opnd1.reg < Opnd2.reg); // no subword regs to deal with

			default: msg("ERROR: Unknown operand type in LessDefinition.\n"); return false;
		}; // end switch (Opnd1.type)
#endif
	} // end operator
}; // end class LessDefinition

// MACHINE DEPENDENT: comparison class to permit sorting of stack delta definitions, each of which is a std::pair<STARSOpndTypePtr, std::pair<STARS_ea_t, STARS_sval_t> >.
class LessStackDeltaCopy {
public:
	bool operator()(const std::pair<STARSOpndTypePtr, std::pair<STARS_ea_t, STARS_sval_t> > &DeltaCopy1, const std::pair<STARSOpndTypePtr, std::pair<STARS_ea_t, STARS_sval_t> > &DeltaCopy2) const {
		STARSOpndTypePtr Opnd1 = DeltaCopy1.first;
		STARSOpndTypePtr Opnd2 = DeltaCopy2.first;
		if (IsEqOp(Opnd1, Opnd2)) {
			// If definitions are of the same operand, sort by instruction address of the definition.
			return (DeltaCopy1.second.first < DeltaCopy2.second.first);
		}
		else if (Opnd1->IsRegOp() && Opnd2->IsRegOp()) {
			return MDLessRegOpnd(Opnd1, Opnd2);
		}
#if 1
		return ((*Opnd1) < (*Opnd2));  // use < operator from STARS_op_t
#else
		if (Opnd1->GetOpType() != Opnd2->GetOpType())
			return (Opnd1->GetOpType() < Opnd2->GetOpType());
		switch (Opnd1->GetOpType()) {
			case o_void: return false;
			case o_reg: return MDLessRegOpnd(Opnd1, Opnd2);
			case o_mem: return (Opnd1->GetAddr() < Opnd2->GetAddr());
			case o_phrase: if (Opnd1->HasSIBByte() && Opnd2->HasSIBByte()) return ((Opnd1->GetSIB() < Opnd2->GetSIB()) || ((Opnd1->GetSIB() == Opnd2->GetSIB()) && (Opnd1->GetSpecFlag4() < Opnd2->GetSpecFlag4())));
							else if (Opnd2->HasSIBByte()) return true;  // no SIB < has SIB
							else if (Opnd1->HasSIBByte()) return false; // no SIB < has SIB
							else return MDLessReg(Opnd1->GetReg(), Opnd2->GetReg()); // no SIB bytes
			case o_displ: if (Opnd1->HasSIBByte() && Opnd2->HasSIBByte())
							  return ((Opnd1->GetSIB() < Opnd2->GetSIB()) 
							  || ((Opnd1->GetSIB() == Opnd2->GetSIB()) && (Opnd1->GetSpecFlag4() < Opnd2->GetSpecFlag4())));
						  else if (Opnd2->HasSIBByte()) return true;  // no SIB < has SIB
						  else if (Opnd1->HasSIBByte()) return false; // no SIB < has SIB
						  else return ((Opnd1->GetAddr() < Opnd2->GetAddr())
								|| ((Opnd1->GetAddr() == Opnd2->GetAddr()) 
								&& MDLessReg(Opnd1->GetReg(), Opnd2->GetReg()))); // no SIB bytes
			case o_imm: return (Opnd1->GetImmedValue() < Opnd2->GetImmedValue());
			case o_far:  // fall through to o_near case
			case o_near: return (Opnd1.addr < Opnd2.addr);
			case o_trreg:  // fall through
			case o_dbreg:  // fall through
			case o_crreg:  // fall through
			case o_fpreg:  // fall through
			case o_mmxreg: // fall through
			case o_xmmreg: // fall through
			case o_ymmreg: return (Opnd1.reg < Opnd2.reg); // no subword regs to deal with

			default: msg("ERROR: Unknown operand type in LessStackDeltaCopy.\n"); return false;
		}; // end switch (Opnd1.type)
#endif
	} // end operator
}; // end class LessStackDeltaCopy

// Get the size in bytes of the data type of an operand.
std::size_t GetOpDataSize(const STARSOpndTypePtr &DataOp);

// Get the IDA Pro register size (dtyp) field
char GetRegDtyp(uint16_t RegNum, bool Has64BitOpnds);

// STARS will operate on a linked list of instructions, which
//  will be grouped into basic blocks. We will include info about each
//  instruction that can not all be obtained easily in IDA Pro, and
//  is not all obtained easily from one place in IDA Pro. This 
//  information will include all that we need to perform data flow 
//  analyses. Because the LABEL and CASE could be applied to other
//  kinds of instructions during analysis, we define the enum values
//  as non-overlapping bits. This type might be stored as an int
//  in the future as a result of this design decision.
enum SMPitype {
	DEFAULT = 0,     // most instructions in middle of basic block
	LABEL = 1,       // instr also has a label (can be jump target) UNUSED
	CASE = 2,        // instr also has case label (is inside switch) UNUSED
	JUMP = 4,        // unconditional branch
	COND_BRANCH = 8, // conditional branch
	INDIR_JUMP = 16, // indirect unconditional branch,
	CALL = 32,       // direct call
	INDIR_CALL = 64, // indirect call
	RETURN = 128,    // function return OR TAIL CALL
	HALT = 256       // execution stops
};

// A primary goal of data flow analysis in SMP will be to identify
//  operands as numeric or pointer type. If an instruction definitely
//  produces a numeric result, an annotation could inform the mmStrata
//  run time monitor to simply record 'numeric' for the result register
//  and otherwise do nothing. This could greatly cut down the overhead
//  at run time. If the result register already has an mmStrata metadata
//  type of numeric, then the annotation could tell mmStrata to do nothing
//  at all, which is even better. In order to emit these annotations, we
//  will track all operands in a function as to their basic SMP types.
#define PROF_BASE 128
enum SMPOperandType { // What type is a given register or memory operand?
	UNINIT = 0,      // Operand type not yet analyzed; type lattice top
	NUMERIC = 1,     // Definitely holds non-pointer value (char, int, etc.)
	CODEPTR = 2,     // Definitely holds a code address; mmStrata considers this NUMERIC
	POINTER = 4,     // Definitely holds a data address
	STACKPTR = 8,    // Definitely holds a stack data address (refinement of POINTER)
	GLOBALPTR = 16,   // Definitely holds a global static data address (refinement of POINTER)
	HEAPPTR = 32,     // Definitely holds a heap data address (refinement of POINTER)
	PTROFFSET = 64,   // Offset from a pointer, e.g. a difference between two pointers
	NEGATEDPTR = 65,  // Temporary ptr arithmetic leaves NUMERIC-POINTER; should get +POINTER later to make PTROFFSET
	UNKNOWN = 96,     // Might hold an address, might not (Bad!); type lattice bottom
	PROF_NUMERIC = NUMERIC + PROF_BASE, // NUMERIC, determined using profiler input
	PROF_CODEPTR = CODEPTR + PROF_BASE, // CODEPTR, determined using profiler input
	PROF_POINTER = POINTER + PROF_BASE, // POINTER, determined using profiler input
	PROF_STACKPTR = STACKPTR + PROF_BASE, // STACKPTR, determined using profiler input
	PROF_GLOBALPTR = GLOBALPTR + PROF_BASE, // GLOBALPTR, determined using profiler input
	PROF_HEAPPTR = HEAPPTR + PROF_BASE, // HEAPPTR, determined using profiler input
	PROF_PTROFFSET = PTROFFSET + PROF_BASE, // PTROFFSET, determined using profiler input
	PROF_NEGATEDPTR = NEGATEDPTR + PROF_BASE, // NEGATEDPTR, determined using profiler input
	PROF_UNKNOWN = UNKNOWN + PROF_BASE // UNKNOWN, determined using profiler input
};

// Macros to query and compare SMPOperandType values.

// Is OpType any kind of pointer to data, whether profile derived or not?
#define IsDataPtr(OpType) (((OpType & (~PROF_BASE)) >= POINTER) && ((OpType & (~PROF_BASE)) <= HEAPPTR))

// Is OpType one of the refined data pointer types (stack, global, heap)?
#define IsRefinedDataPtr(OpType) (((OpType & (~PROF_BASE)) >= STACKPTR) && ((OpType & (~PROF_BASE)) <= HEAPPTR))

// Is OpType the lattice bottom (overdefined) type UNKNOWN, profile derived or not?
#define IsUnknown(OpType) ((OpType & (~PROF_BASE)) == UNKNOWN)

// Is OpType1 the same type as OpType2, ignoring profile derivation?
#define IsEqType(OpType1, OpType2)  ((OpType1 & (~PROF_BASE)) == (OpType2 & (~PROF_BASE)))
#define IsNotEqType(OpType1, OpType2)  ((OpType1 & (~PROF_BASE)) != (OpType2 & (~PROF_BASE)))

// Is OpType some kind of NUMERIC (i.e. NUMERIC or CODEPTR)?
#define IsNumeric(OpType) (IsEqType(OpType, NUMERIC) || IsEqType(OpType, CODEPTR))

// Is OpType a profile derived type?
#define IsProfDerived(OpType)  (0 != (OpType & PROF_BASE))

// Make OpType into its equivalent profile derived type.
#define MakeProfDerived(OpType) ((SMPOperandType)(OpType | PROF_BASE))

// Meet function over any two types in the type lattice.
SMPOperandType SMPTypeMeet(SMPOperandType Type1, SMPOperandType Type2);

// Meet function for SCCP constant propagation; updates NewConstStruct
void STARSConstantTypeMeet(struct STARS_SCCP_Const_Struct OldConstStruct, struct STARS_SCCP_Const_Struct &NewConstStruct);

// Enumeration to keep track of whether the metadata for each DEF is
//  used, unused, redundant, etc.
enum SMPMetadataType {
	DEF_METADATA_UNANALYZED = 0, // have not finished analyzing
	DEF_METADATA_UNUSED = 1,   // will not be used in a later store inst
	DEF_METADATA_USED = 2,     // will be used in bounds check for a store
	DEF_METADATA_USED_MEMWRITE = 4, // explanation of why DEF_METADATA_USED bit is set; reaches memory write
	DEF_METADATA_USED_CALL = 8, // explanation of why DEF_METADATA_USED bit is set; reaches call
	DEF_METADATA_USED_RETURN = 16, // explanation of why DEF_METADATA_USED bit is set; reaches return
	DEF_METADATA_REDUNDANT = 32, // DEF reg already has the same metadata
								// it would receive by updating from this inst
	DEF_METADATA_PROF_REDUNDANT = 64  // redundant, based on profiler-derived types
};

// Enumeration for constant propagation type lattice.
enum STARSConstantValueType {
	STARS_CONST_TOP = 0,  // top of lattice, uninitialized
	STARS_CONST_HAS_VALUE = 1, // Found 1 constant value so far; value is in SMPFunction::ConstantDefs
	STARS_CONST_BOTTOM = 2  // bottom of lattice, more than one value possible
};

struct STARS_SCCP_Const_Struct {
	enum STARSConstantValueType ConstType;
	STARS_uval_t ConstValue;   // holds valid value only if ConstType == STARS_CONST_HAS_VALUE.
};

// Encoding of info for fine-grained analysis of stack frames, integer error analysis, etc.
struct FineGrainedInfo {
	unsigned short SizeInfo;  // 8-bit, 16-bit, ...128-bit, int, float, string
	unsigned short SignMiscInfo; // signed, unsigned, read, written, FP-relative, etc
};

// Masks for FineGrainedInfo.SignMiscInfo
#define FG_MASK_SIGNED 1
#define FG_MASK_UNSIGNED 2
#define FG_MASK_INCONSISTENT_SIGN 3
#define FG_MASK_SIGNEDNESS_BITS 3  // better name for some uses
#define FG_MASK_READ 4
#define FG_MASK_WRITTEN 8
#define FG_MASK_SP_RELATIVE 16
#define FG_MASK_FP_RELATIVE 32
#define FG_MASK_ADDRESS_TAKEN 64
#define FG_MASK_INDEXED_ACCESS 128

// Masks for FineGrainedInfo.SizeInfo
#define FG_MASK_INTEGER 1
#define FG_MASK_STRING 2  // string data
#define FG_MASK_FLOAT_MMX 4  // floating point or MMX/SSE/packed
// remaining bit fields are the bit widths, e.g. 8 => 8-bit, ... 128 => 128-bit
#define FG_MASK_BITWIDTH_8 8
#define FG_MASK_BITWIDTH_16 16
#define FG_MASK_BITWIDTH_32 32
#define FG_MASK_BITWIDTH_64 64
#define FG_MASK_BITWIDTH_128 128
#define FG_MASK_BITWIDTH_256 256
#define FG_MASK_BITWIDTH_80 80   // two bits are set here
#define FG_MASK_BITWIDTH_96 512
#define FG_MASK_BITWIDTH_48 1024
#define FG_MASK_BITWIDTH_24 2048
#define FG_MASK_BITWIDTH_FIELDS (8|16|32|64|128|256|512|1024|2048)
// miscellaneous type info
#define FG_MASK_CODEPOINTER 4096
#define FG_MASK_DATAPOINTER 8192

// We don't worry about truncation unless we see operands less than normal width.
//  This would change if we port to 64-bit machines.
#define MD_NORMAL_MACHINE_BITWIDTH 32
#define MD_NORMAL_BITWIDTH_MASK FG_MASK_BITWIDTH_32

// Return one of the bit width masks above for the current operand.
unsigned short ComputeOperandBitWidthMask(const STARSOpndTypePtr &CurrOp, std::size_t DataSize);

// Compute largest bit width from a SignMiscInfo bit mask.
std::size_t LargestBitWidthFromMask(unsigned short WidthTypeInfo);

// Map the two sign bits above into four strings for annotation output.
extern const char *SignednessStrings[4];
// Version with NOFLAG prefixed for the load effective address opcodes
//  that can overflow silently without setting any flags.
extern const char *LeaSignednessStrings[4];

class DisAsmString {
public:
	DisAsmString(void);
	char *GetDisAsm(STARS_ea_t InstAddr, bool MarkerInst = false);
	void SetMarkerInstText(STARS_ea_t InstAddr);

private:
	STARS_ea_t CurrAddr; // cached string corresponds to this instruction address
	STARS_ssize_t StringLen; // string length returned by IDA Pro
	char CachedDisAsm[MAXSMPDISASMSTR];
};

extern DisAsmString DisAsmText;

// Masks to set bits to use as booleans1 in DefOrUse.
#define DEFORUSE_SET_INDIRECT_WRITE 0x01
#define DEFORUSE_SET_NO_TRUNCATION 0x02
#define DEFORUSE_SET_NO_OVERFLOW 0x04
#define DEFORUSE_SET_GLOBALLY_LOOP_INVARIANT 0x08
#define DEFORUSE_SET_UNUSED3 0x10
#define DEFORUSE_SET_UNUSED4 0x20
#define DEFORUSE_SET_UNUSED5 0x40
#define DEFORUSE_SET_UNUSED6 0x80
#define DEFORUSE_RESET_INDIRECT_WRITE 0xfe
#define DEFORUSE_RESET_NO_TRUNCATION 0xfd
#define DEFORUSE_RESET_NO_OVERFLOW 0xfb
#define DEFORUSE_RESET_GLOBALLY_LOOP_INVARIANT 0xf7
#define DEFORUSE_RESET_UNUSED3 0xef
#define DEFORUSE_RESET_UNUSED4 0xdf
#define DEFORUSE_RESET_UNUSED5 0xbf
#define DEFORUSE_RESET_UNUSED6 0x7f

class DefOrUse {
public:
	// Constructors
	DefOrUse(void);
	DefOrUse(STARSOpndTypePtr Ref, SMPOperandType Type = UNINIT, int SSASub = SMP_SSA_UNINIT);
	DefOrUse(const DefOrUse &CopyIn);

	// Operators
	DefOrUse &operator=(const DefOrUse &rhs);

	// Get methods
	inline STARSOpndTypePtr GetOp(void) const { return Operand; };
	inline SMPOperandType GetType(void) const { return OpType; };
	inline int GetSSANum(void) const { return SSANumber; };
	inline SMPMetadataType GetMetadataStatus(void) const { return MetadataStatus; };

	// Query methods.
	inline bool HasIndirectWrite(void) const { return (booleans1 & DEFORUSE_SET_INDIRECT_WRITE); }; // DEF has ind. write before its last USE
	inline bool DoesNotTruncate(void) const { return (booleans1 & DEFORUSE_SET_NO_TRUNCATION); };
	inline bool DoesNotOverflow(void) const { return (booleans1 & DEFORUSE_SET_NO_OVERFLOW); };
	inline bool IsInvariantForAllLoops(void) const { return (booleans1 & DEFORUSE_SET_GLOBALLY_LOOP_INVARIANT); };

	// Set methods
	inline void SetOp(STARSOpndTypePtr NewOp) { Operand = NewOp; };
	inline void SetSSANum(int Num) { SSANumber = Num; };
	void SetType(SMPOperandType Type, const SMPInstr* instr); 
	void SetMetadataStatus(SMPMetadataType NewStatus);
	inline void SetIndWrite(bool IndMemWrite) { IndMemWrite ? (booleans1 |= DEFORUSE_SET_INDIRECT_WRITE) : (booleans1 &= DEFORUSE_RESET_INDIRECT_WRITE); };
	inline void SetNoTruncation(bool NoTruncFlag)  { NoTruncFlag ? (booleans1 |= DEFORUSE_SET_NO_TRUNCATION) : (booleans1 &= DEFORUSE_RESET_NO_TRUNCATION); };
	inline void SetNoOverflow(bool NoOverflowFlag)  { NoOverflowFlag ? (booleans1 |= DEFORUSE_SET_NO_OVERFLOW) : (booleans1 &= DEFORUSE_RESET_NO_OVERFLOW); };
	inline void SetInvariantForAllLoops(void)  { booleans1 |= DEFORUSE_SET_GLOBALLY_LOOP_INVARIANT; };

	// Printing methods
	void Dump(void) const;
private:
	// Data
	STARSOpndTypePtr Operand;
	int SSANumber;
	SMPOperandType OpType;
	SMPOperandType NonSpeculativeOpType;
	SMPMetadataType MetadataStatus;
	unsigned char booleans1; // holds bit-flags
#if 0  // Moved flags into booleans1.
	bool IndWrite; // Does an indirect memory write occur within the chain
	               //  starting at this DEF? (always false for a USE)
	bool NoTruncation; // USE appears to be a truncation of its original DEF, but is not.
	bool NoOverflow; // DEF appears to be an overflowing computation, but is not.
	bool GloballyLoopInvariant; // DEF is not variant in any loop
#endif
}; // end of class DefOrUse

// Comparison operator class to permit use of DefOrUse type in sets.
class LessDefUse {
public:
	bool operator()(const DefOrUse &Ref1, const DefOrUse &Ref2) const {
		std::set<STARSOpndTypePtr, LessOp>::value_compare OpComp = this->DummySet.value_comp(); // LessOp()
		return OpComp(Ref1.GetOp(), Ref2.GetOp());
	}
private:
	std::set<STARSOpndTypePtr, LessOp> DummySet;
};

typedef std::set<DefOrUse, LessDefUse> STARSDefUseSet;
typedef std::set<DefOrUse, LessDefUse>::iterator STARSDefUseIter;

// Same class is used for both a DEF set and a USE set.
class DefOrUseSet {
public:
	// Constructors
	DefOrUseSet(void);
	// Destructor.
	~DefOrUseSet();

	// Get methods
	inline std::size_t GetSize(void) const { return (std::size_t) Refs.size(); };
	inline std::set<DefOrUse, LessDefUse>::iterator GetFirstRef(void) { return Refs.begin(); };
	inline std::set<DefOrUse, LessDefUse>::iterator GetLastRef(void) { return Refs.end(); };
	std::set<DefOrUse, LessDefUse>::iterator FindRef(const STARSOpndTypePtr &SearchOp);
	
	// Set methods
	// Insert a new DEF or USE; must be new, insert must succeed else we assert.
	std::set<DefOrUse, LessDefUse>::iterator InsertRef(DefOrUse Ref); 
	void SetRef(STARSOpndTypePtr Ref, SMPOperandType Type = UNINIT, int SSASub = SMP_SSA_UNINIT);
	std::set<DefOrUse, LessDefUse>::iterator SetOp(std::set<DefOrUse, LessDefUse>::iterator CurrRef, STARSOpndTypePtr NewOp);
	std::set<DefOrUse, LessDefUse>::iterator SetSSANum(const STARSOpndTypePtr &CurrOp, int NewSSASub);
	std::set<DefOrUse, LessDefUse>::iterator SetType(const STARSOpndTypePtr &CurrOp, SMPOperandType Type, const SMPInstr* Instr);
	std::set<DefOrUse, LessDefUse>::iterator SetMetadata(const STARSOpndTypePtr &CurrOp, SMPMetadataType Status);
	// Change the indirect write status for a reference:
	std::set<DefOrUse, LessDefUse>::iterator SetIndWrite(const STARSOpndTypePtr &CurrOp, bool IndWriteFlag);
	// Change the ignore apparent truncation flag for a reference:
	std::set<DefOrUse, LessDefUse>::iterator SetNoTruncation(const STARSOpndTypePtr &CurrOp, bool NoTruncFlag);
	// Change the ignore apparent overflow flag for a reference:
	std::set<DefOrUse, LessDefUse>::iterator SetNoOverflow(const STARSOpndTypePtr &CurrOp, bool NoOverflowFlag);
	// Mark a DEF as invariant for all loops in the func:
	std::set<DefOrUse, LessDefUse>::iterator SetLoopInvariant(const STARSOpndTypePtr &CurrOp);

	inline void clear(void) { Refs.clear(); return; };
	inline void EraseRef(std::set<DefOrUse, LessDefUse>::iterator RefIter) {
		Refs.erase(RefIter); return;
	};

	// Printing methods
	void Dump(void) const;

	// Analysis methods
	bool TypesAgreeNoFlags(void); // Are all types consistent, ignoring flag registers?
private:
	// Data
	std::set<DefOrUse, LessDefUse> Refs;  // Defined or used operand with type and SSA subscript
};  // end class DefOrUseSet

// Same class is used for both a DEF list and a USE list.
class DefOrUseList {
public:
	// Constructors
	DefOrUseList(void);

	// Get methods
	DefOrUse GetRef(std::size_t index) const;
	inline std::size_t GetSize(void) const { return (std::size_t) Refs.size(); };
	inline DefOrUse *GetRefNum(std::size_t index) { return &Refs[index]; };
	inline std::vector<DefOrUse>::iterator GetFirstRef(void) { return Refs.begin(); };
	inline std::vector<DefOrUse>::iterator GetLastRef(void) { return Refs.end(); };
	inline int GetRefSSANum(std::size_t index) const { return Refs.at(index).GetSSANum(); };
	inline SMPOperandType GetRefType(std::size_t index) const { return Refs.at(index).GetType(); };

	// Set methods
	void SetRef(STARSOpndTypePtr Ref, SMPOperandType Type = UNINIT, int SSASub = SMP_SSA_UNINIT);
	void SetSSANum(std::size_t index, int NewSSASub);
	void SetType(std::size_t index, SMPOperandType Type, const SMPInstr* Instr);
	inline void clear(void) { Refs.clear(); return; };
	// Printing methods
	void Dump(void) const;
	// Analysis methods
	void EraseDuplicates(void);  // in case SMPInstr::MDFixupDefUseLists() adds duplicate
private:
	// Data
	std::vector<DefOrUse> Refs;  // Defined or used operand with type, SSA subscript, etc.
};  // end class DefOrUseList

class SMPPhiFunction {
public:
	// Constructors
	SMPPhiFunction(int GlobIndex, const DefOrUse &Def);

	// Get methods
	inline int GetIndex(void) const { return index; };
	inline std::size_t GetPhiListSize(void) const { return this->SubscriptedOps.GetSize(); };
	DefOrUse GetDefCopy(void) const;
	inline DefOrUse GetPhiRef(std::size_t i) const { return this->SubscriptedOps.GetRef(i); };
	inline DefOrUse *GetRefNum(std::size_t index) { return SubscriptedOps.GetRefNum(index); };
	inline std::vector<DefOrUse>::iterator GetFirstOp(void) { return SubscriptedOps.GetFirstRef(); };
	inline std::vector<DefOrUse>::iterator GetLastOp(void) { return SubscriptedOps.GetLastRef(); };
	inline STARSOpndTypePtr GetAnyOp(void) const { return DefName.GetOp(); };
	inline int GetDefSSANum(void) const { return DefName.GetSSANum(); };
	inline int GetUseSSANum(std::size_t index) const { return SubscriptedOps.GetRefSSANum(index); };
	inline SMPOperandType GetDefType(void) const { return DefName.GetType(); };
	inline SMPOperandType GetUseType(std::size_t index) const { return SubscriptedOps.GetRefType(index); };
	inline SMPMetadataType GetDefMetadata(void) const { return DefName.GetMetadataStatus(); };

	// Set methods
	void PushBack(DefOrUse Ref); // add inputs to phi function
	void SetSSADef(int NewSSASub);
	void SetSSARef(std::size_t index, int NewSSASub);
	void SetDefType(SMPOperandType Type, const SMPInstr* instr);
	void SetRefType(std::size_t index, SMPOperandType Type, const SMPInstr* instr);
	void SetDefMetadata(SMPMetadataType Status);

	// Query methods
	bool HasTypedUses(void); // true ==> at least one USE is not UNINIT type
	// Printing methods
	void Dump(void) const;
	// Analysis methods
	SMPOperandType ConditionalMeetType(SMPBasicBlock *CurrBlock) const; // meet operator over USE types
		// WARNING: This method propagates types, which can cause the Phi iterator in the enclosing SMPBasicBlock to change
		//  and no longer be equal to the "this" pointer in this SMPPhiFunction. Need to re-do FindPhi() to get non-stale iterator.

private:
	int index;  // index into SMPFunction::BlocksDefinedIn (i.e. global name index)
	DefOrUse DefName; // output of phi function
	DefOrUseList SubscriptedOps; // ordered list of inputs to phi function
}; // end class SMPPhiFunction

// Comparison function to permit sorting phi functions and keeping them in sets.
class LessPhi {
public:
	bool operator() (const SMPPhiFunction &Phi1, const SMPPhiFunction &Phi2) const {
		return (Phi1.GetIndex() < Phi2.GetIndex());
	}
}; // end of class LessPhi

// In SSA computations, we are storing the GlobalNames index into the op_t fields
//  n, offb, and offo. This function extracts an unsigned int from these three 8-bit
//  fields. The GlobalNames index is only stored in the GlobalNames set, although
//  they might be found (unreliably) elsewhere because STARSOpndTypePtr is a shared_ptr type.
unsigned int ExtractGlobalIndex(const STARSOpndTypePtr &GlobalOp);
void SetGlobalIndex(STARSOpndTypePtr TempOp, std::size_t index);

class SMPDefUseChain {
public:
	// Constructors
	SMPDefUseChain(void);
	SMPDefUseChain(STARSOpndTypePtr Name, STARS_ea_t Def = STARS_BADADDR);
	// Get methods
	inline STARSOpndTypePtr GetName(void) const { return SSAName; };
	inline STARS_ea_t GetDef(void) const { return (STARS_ea_t) RefInstrs.at(0); };
	inline STARS_ea_t GetUse(std::size_t index) const { return (STARS_ea_t) RefInstrs.at(index + 1); };
	inline STARS_ea_t GetLastUse(void) const {
		if (RefInstrs.size() > 1) 
			return (STARS_ea_t) RefInstrs.back();
		else
			return (STARS_ea_t) STARS_BADADDR;
	};
	inline std::size_t GetNumUses(void) const { return RefInstrs.size() - 1; };

	// Query methods.
	inline bool HasIndirectWrite(void) const { return IndWrite; };

	// Set methods
	void SetName(STARSOpndTypePtr Name);
	void SetDef(STARS_ea_t Def);
	void PushUse(STARS_ea_t Use);
	void SetIndWrite(bool IndMemWrite);
	// Printing methods
	void Dump(int SSANum = (-1)) const;
private:
	STARSOpndTypePtr SSAName;  // What variable is defined and used in the chain?
	bool IndWrite; // Does an indirect memory write occur within the chain?

	// Keep DEF and USE addrs as unsigned short offsets from BaseAddr of block.
	//  The BaseAddr of the block is 1 less than the first inst addr of the block,
	//  so that this value can be used as a flag to indicate that the value was
	//  live-in and upward-exposed upon block entry and does not have a real DEF
	//  in the block for that SSANum.
	std::vector<unsigned short> RefInstrs; // First is always DEF, rest are USE.
}; // end class SMPDefUseChain

class SMPDUChainArray {
public:
	// Constructor
	SMPDUChainArray(void);
	SMPDUChainArray(STARSOpndTypePtr Name, STARS_ea_t FirstAddrMinusOne);
	// Get methods.
	inline STARSOpndTypePtr GetName(void) const { return SSAName; };
	inline STARS_ea_t GetDef(int SSANum) const { return BaseAddr + DUChains.at(SSANum).GetDef(); };
	inline STARS_ea_t GetUse(int SSANum, std::size_t index) const { return BaseAddr + DUChains.at(SSANum).GetUse(index); };
	STARS_ea_t GetLastUse(int SSANum) const;
	inline std::size_t GetNumUses(int SSANum) const { return DUChains.at(SSANum).GetNumUses(); };
	inline std::size_t GetSize(void) const { return DUChains.size(); };
	// Set methods.
	void SetName(STARSOpndTypePtr Name, STARS_ea_t FirstAddrMinusOne);
	inline void PushUse(int SSANum, STARS_ea_t UseAddr) { DUChains.at(SSANum).PushUse(UseAddr - BaseAddr); };
	inline void PushChain(SMPDefUseChain NewChain) { DUChains.push_back(NewChain); };
	inline void SetIndWrite(int SSANum, bool NewFlag) { DUChains.at(SSANum).SetIndWrite(NewFlag); };
	// Query methods
	inline bool HasIndirectWrite(int SSANum) const { return DUChains.at(SSANum).HasIndirectWrite(); };
	// Printing methods.
	void Dump(void) const;
private:
	// Data
	std::vector<SMPDefUseChain> DUChains; // indexed by SSA number for local chains, by chain # for global chains
	STARSOpndTypePtr SSAName; // What variable is used in all chains in the array?
	STARS_ea_t BaseAddr; // one less than first addr in basic block
}; // end class SMPDUChainArray

class SMPCompleteDUChains {
public:
	// Printing methods.
	void Dump(void) const;
	// Data (public for convenience)
	std::vector<SMPDUChainArray> ChainsByName; // indexed by name index
}; // end class SMPCompleteDUChains

// Our own bitset class. The C++ bitset class cannot be dynamically sized. The vector<bool> specialization
//  of the C++ STL vector container does not behave like other containers and is in danger of being deprecated.
class STARSBitSet {
public:
	// Constructors
	STARSBitSet();
	// Get methods
	bool GetBit(std::size_t BitIndex) const;
	// Set methods
	void AllocateBits(std::size_t Size); // allocate Size bits, initialized to zero.
	void SetBit(std::size_t BitIndex);
	void ResetBit(std::size_t BitIndex);
	// Query methods
	bool IsAnyBitSet(void) const; // Returns false if all bits are zero, true otherwise.
	int FindHighestBitSet(void) const; // Return highest index set; -1 if no bits set.

private:
	// Data members.
	std::size_t BitLimit; // number of bits
	std::vector<unsigned char> STARSBits;
	// Methods
}; // end class STARSBitSet

// Initialization routine for DFA category.
extern SMPitype DFACategory[STARS_NN_last + 1];

// Initialization routine for Type category.
extern int SMPTypeCategory[STARS_NN_last + 1];

// Initializations for CPU flags DEF/USE.
extern bool SMPDefsFlags[STARS_NN_last + 1];
extern bool SMPUsesFlags[STARS_NN_last + 1];

// Initialize the FG info for the return register from any library function
//  whose name implies that we know certain return values (e.g. atoi() returns
//  a signed integer, while strtoul() returns an unsigned long).
void GetLibFuncFGInfo(std::string FuncName, struct FineGrainedInfo &InitFGInfo);

// Map system or library call name to FG info about its return value.
extern std::map<std::string, struct FineGrainedInfo> ReturnRegisterTypeMap;

// Init map of system or library call name to the annotation substring that
//  guides saturating arithmetic or other continuation policies in 
//  the case of integer error detection of a value passed to that call.
void InitIntegerErrorCallSinkMap(void);

// Return sink string for call name from the sink map.
// If we don't care find the call name, we return an empty string.
void GetSinkStringForCallName(std::string CalleeName, std::string &SinkString);

// Init map of system or library call name to the argument number that
//  should have an unsigned value and should be guarded from the
//  signedness error that results from copying a signed value
//  into the outgoing argument. Argument numbers are zero-based; see bit defs below.
void InitUnsignedArgPositionMap(void);

// Define argument position bits to put into various Arg Position Maps.
// A system or library call name could have more than one argument
//  position to protect from signedness errors.
#define STARS_ARG_POS_0 0x1
#define STARS_ARG_POS_1 0x2
#define STARS_ARG_POS_2 0x4
#define STARS_ARG_POS_3 0x8
#define STARS_ARG_POS_4 0x16

// Return unsigned argument position bitmap for call name from the unsigned arg position map.
//  We will return 0 when there is no argument to worry about
//  for a particular library or system call name.
void GetUnsignedArgPositionsForCallName(std::string CalleeName, unsigned int &ArgPosBits);

// Init map of system or library call name to the argument number that is dangerous
//  to accept from a user-tainted source. Lots of overlap with UnsignedArgPositionMap.
void InitTaintWarningArgPositionMap(void);

// Return unsigned argument position bitmap for call name from the taint warning arg position map.
//  We will return 0 when there is no argument to worry about
//  for a particular library or system call name.
void GetTaintWarningArgPositionsForCallName(std::string CalleeName, unsigned int &ArgPosBits);

// Init map of system or library call name to the argument number that
//  should have a POINTER value.
void InitPointerArgPositionMap(void);

// Return POINTER argument position bitmap for call name from the taint warning arg position map.
//  We will return 0 when there is no POINTER argument to worry about
//  for a particular library or system call name.
void GetPointerArgPositionsForCallName(std::string CalleeName, unsigned int &ArgPosBits);

// Utility to count bits set in an unsigned int, e.g. ArgPosBits.
unsigned int CountBitsSet(unsigned int ArgPosBits);

// Utility to get highest bit set in a byte.
unsigned int HighestBitSet(unsigned char Byte);

// Is CalleeName a standard library function name?
bool IsLibFuncName(std::string CalleeName);

// Is FuncName a startup func called before main(), or a wrapup function called by the system?
bool IsStartupFuncName(const std::string FuncName);

#endif
