#ifndef STARS_IDA_op_h
#define STARS_IDA_op_h

#include <stdint.h>

#include <pro.h>
#include <ua.hpp>
#include <intel.hpp>

#include "interfaces/STARSTypes.h"
#include "base/SMPDataFlowAnalysis.h"

class STARS_IDA_op_t : public STARS_op_t
{
	public:
		
		// Constructors
		STARS_IDA_op_t(op_t IDAOp);

		// Clone method (deep copy).
		STARSOpndTypePtr clone(void) const;

		// Operators
		bool operator<(const STARS_op_t &rOp) const; // Less-than operator for use in STL ordered containers, e.g. sets.

		// Get methods
		unsigned int GetOpGlobalIndex(void) const; // Get STARS SSA name index.
		STARS_ea_t GetAddr(void) const { return m_Opnd.addr; };
		unsigned char GetOpType(void) const { return m_Opnd.type; };
		uint16_t GetReg(void) const { return m_Opnd.reg; };
		uint16_t GetSegReg(void) const { return m_Opnd.segrg; }; // Get segment register of operand; R_none if no segment register
		char GetSIB(void) const { return m_Opnd.sib; }; // Get x86 SIB byte (dense encoding of base reg, index reg, and scale factor)
		int MDGetSIBBaseReg(void) const; // Get x86 SIB byte base register
		short MDGetSIBIndexReg(void) const; // Get x86 SIB byte index register
		int GetSIBScaleFactor(void) const { return ((m_Opnd.sib >> 6) & 3); }; // Extract scale factor from SIB byte
		char GetSpecFlag4(void) const { return m_Opnd.specflag4; }; // Get specflag4 byte, used to hold a copy of the auxpref byte in x86-64 programs.
		STARS_uval_t GetImmedValue(void) const { return m_Opnd.value; }; // Get value field for immediate operands; uint32 for x86-32, uint64 for x86-64
		char GetOpDtyp(void) const { return m_Opnd.dtyp; };  // Get field that determines byte width
		uint16_t GetByteWidth(void) const;

		// Set methods
		void SetByteWidth(uint16_t ByteWidth); // Change default byte width of opnd
		void DoubleRegWidth(void); // Double the width of opnd if it is a subword register
		void SetOpGlobalIndex(std::size_t index); // Set STARS SSA name index.
		void SetReg(uint16_t NewReg);  // Change the register field of the operand.
		void SetAddr(STARS_ea_t NewAddr) { m_Opnd.addr = NewAddr; };
		void SetSIB(char value) { m_Opnd.sib = value; };
		void SetSpecFlag4(char value) { m_Opnd.specflag4 = value; };
		void SetBitInSpecFlag4(char value) { m_Opnd.specflag4 |= value; };
		void SetNotVisible(void) { // Set operand to type that is implicit, not explicit in disasm
#if IDA_SDK_VERSION < 680
			m_Opnd.clr_showed();	
#else
			m_Opnd.clr_shown();
#endif
		}; 
		void SetTypeToMemDisplacement(void) { m_Opnd.type = o_displ; }; // change type during stack normalization

		// Query methods
		// Operand appears in disasm, vs. being implicit, e.g. "imul ecx" also has eax opnd
		bool IsVisible(void) const { 
#if IDA_SDK_VERSION < 680
			return m_Opnd.showed();
#else
			return m_Opnd.shown();
#endif
		};
		bool IsRegOp(void) const { return (m_Opnd.type == o_reg); };
		bool IsImmedOp(void) const { return (m_Opnd.type == o_imm); };
		bool IsVoidOp(void) const { return (m_Opnd.type == o_void); };
		bool IsMemDisplacementOp(void) const { return (m_Opnd.type == o_displ); };
		bool IsStaticMemOp(void) const { return (m_Opnd.type == o_mem); };
		bool IsMemNoDisplacementOp(void) const { return (m_Opnd.type == o_phrase); };
		bool IsMemOp(void) const { return (IsStaticMemOp() || IsMemDisplacementOp() || IsMemNoDisplacementOp()); };  // any of the previous three types
		bool HasSIBByte(void) const { return (m_Opnd.hasSIB != 0); };
		bool MatchesReg(uint16_t RegNum) const { return m_Opnd.is_reg((int) RegNum); };
		bool IsNearPointer(void) const { return (m_Opnd.type == o_near); };
		bool IsFarPointer(void) const { return (m_Opnd.type == o_far); };
		bool IsFloatingPointRegOp(void) const { return (m_Opnd.type == o_fpreg); };
		bool IsMMXRegOp(void) const { return (m_Opnd.type == o_mmxreg); };
		bool IsXMMRegOp(void) const { return (m_Opnd.type == o_xmmreg); };
		bool IsYMMRegOp(void) const { return (m_Opnd.type == o_ymmreg); };
		bool IsTestRegOp(void) const { return (m_Opnd.type == o_trreg); };
		bool IsDebugRegOp(void) const { return (m_Opnd.type == o_dbreg); };
		bool IsControlRegOp(void) const { return (m_Opnd.type == o_crreg); };
		bool MDIsKnownOpType(void) const { return ((m_Opnd.type >= o_reg) && (m_Opnd.type <= o_ymmreg)); };
		bool MDIsSpecialRegOpType(void) const { return ((m_Opnd.type >= o_trreg) && (m_Opnd.type <= o_ymmreg)); };
		bool HasSegReg(void) const { return is_segreg(GetSegReg()); }; // Has a segment register

	protected:
		op_t m_Opnd;

	private:
		bool MD_STARS_op256(void) const;        // is VEX.L set?
		bool MD_STARS_is_vsib(void) const;  // does instruction use VSIB variant of the SIB byte?

};

#endif



