#ifndef STARS_irdb_Interface_h
#define STARS_irdb_Interface_h


#include "interfaces/abstract/STARSInterface.h"
#include "interfaces/abstract/STARSInstructionID.h"
#include "interfaces/irdb/STARSSegment.h"
#include "interfaces/irdb/STARSFunction.h"
#include "interfaces/irdb/STARSFunction.h"
#include <elfio/elfio.hpp>
#include <elfio/elfio_dump.hpp>

#include <libIRDB-core.hpp>
#include <libIRDB-util.hpp>
#include <assert.h>
#include <stdarg.h>
#include <stdio.h>
#include <pqxx/pqxx>

class STARS_IRDB_Function_t;

class STARS_IRDB_Interface_t: public STARS_Interface_t
{
public:

// Constructors
	STARS_IRDB_Interface_t(libIRDB::FileIR_t* p_firp, libIRDB::pqxxDB_t &p_pqxx_interface) : 
		firp(p_firp), pqxx_interface(p_pqxx_interface) 
	{ 
		InitSegments();
		InitFunctions();
		// init instruction predecessors.
		instruction_preds.AddFile(p_firp);
		InitIBTAMaps();
	}

	~STARS_IRDB_Interface_t()
	{
		for(int j=0;j<segments.size();j++)
		{
			delete segments[j];
			segments[j]=0;
		}
		for(int j=0;j<functions.size();j++)
		{
			delete functions[j];
			functions[j]=0;
		}
		delete elfiop;
	}

	int GetArchitectureBitWidth() { return  firp->GetArchitectureBitWidth(); }
		
// Segment accessors


	// get the segment that contains the address.
	virtual STARS_Segment_t *getseg(const STARS_ea_t &addr) 
	{ 
		for(int i=0;i<segments.size();++i)
		{
			if(segments[i]==NULL)
				continue;
			if(segments[i]->get_startEA() <= addr && addr < segments[i]->get_endEA())
				return segments[i];
		} 
		return NULL;
	}
	// get the Nth segment
	virtual STARS_Segment_t *getnseg(const int &index) { return segments[index]; }

	// get the number of segments.
	virtual std::size_t get_segm_qty() { return segments.size(); }

	// get the next segment after addr.
	virtual STARS_Segment_t *get_next_seg(const STARS_ea_t &addr) 
	{ 
		if(addr < segments[0]->get_startEA())
			return segments[0];
		for(int i=0;i<segments.size();++i)
		{
			// found seg?
			if(segments[i]->get_startEA() <= addr && addr < segments[i]->get_endEA())
			{
				if(i+1 <= segments.size())
					// return next seg.
					return segments[i+1];
				return NULL;
			}
		}
		return NULL;
	}

// Function accessors

    	// find out how many functions there are
	virtual std::size_t get_func_qty() 
	{  return functions.size(); }  

    	// get the index-th function
	virtual STARS_Function_t *getn_func(int index) 
	{ return functions[index]; }

	// get the function that has the instruction 'instruction_id' as the entry point.
	virtual STARS_Function_t *get_func(STARS_ea_t instruction_id) 
	{ return instr_id_to_func_map[instruction_id]; }; 

	virtual void get_func_name(const STARS_ea_t &ea, char* name, const std::size_t &len) 
	{	
		STARS_IRDB_Function_t* f=instr_id_to_func_map[ea];
		assert(f);
		f->GetFunctionName(name,len);
		return;
	}

// Instruction creation.
	virtual STARS_Instruction_t *CreateInst(STARS_InstructionID_t InstID) ;

// File methods.
	virtual FILE *STARS_fopen(const char *file, const char *mode) 
	{ return fopen(file,mode); }
	virtual int STARS_fclose(FILE *fp)
	{ return fclose(fp); }
	virtual int STARS_msg(const char *format, ...)
	{ va_list ap; va_start(ap,format); return vfprintf(stderr, format, ap); }
	virtual int STARS_fprintf(FILE *fp, const char *format, ...)
	{ va_list ap; va_start(ap,format); return vfprintf(fp, format, ap); }
	virtual int STARS_fscanf(FILE *fp, const char *format, ...)
	{ va_list ap; va_start(ap,format); return vfscanf(fp, format, ap); }
	virtual long STARS_ftell(FILE *fp)
	{ return ftell(fp); }
	virtual char *STARS_fgets(char *buffer, int buflen, FILE *fp)
	{ return fgets(buffer,buflen,fp); }
	virtual int STARS_fseek(FILE *fp, long offset, int whence)
	{ return fseek(fp,offset,whence); }
	virtual int STARS_fgetc(FILE *fp)
	{ return fgetc(fp); }

// String methods
	virtual char *STARS_strncat(char *dst, const char *src, std::size_t dstsize) 
	{ return strncat(dst,src,dstsize); }
	virtual char *STARS_strncpy(char *dst, const char *src, std::size_t dstsize) 
	{ return strncpy(dst,src,dstsize); }
	virtual int STARS_snprintf(char *buffer, std::size_t n, const char *format, ...)
	{ va_list ap; va_start(ap,format); return vsnprintf(buffer,n, format, ap); }

// Disasm methods.
	virtual STARS_ssize_t STARS_tag_remove(const char *instr, char *buf, std::size_t bufsize) 
	{ 
		/* remove comments, etc. from disassembly.*/  
		// do strcpy from instr to buf 
		if(instr != buf)
			strncpy(buf,instr, bufsize);
		return 0; // success
	}

	virtual bool STARS_generate_disasm_line(STARS_ea_t addr, char *buf, std::size_t bufsize, int flags = 0);

// Miscellaneous methods.
	virtual bool InstHasNoCodeXrefs(STARS_InstructionID_t InstID) const
	{ assert(0); return 0; } // no impl'd, shouldn't be called.
	
	virtual bool IsInstJumpTarget(STARS_InstructionID_t InstID) const;

// Find call target; first one, if any, for indirect call
	virtual STARS_InstructionID_t FindFirstCallTarget(STARS_InstructionID_t CallInstID) const;
#if 0
	{ 
		
		// direct call: if no target -- return STARS_BADADDR
		// direct call: if target, return target 
		// indirect call: if any targets exist return one.
		// indirect call: if no known targets, return  STARS_BADADDR ?
		assert(0); return 0; 
	}
#endif

	// Miscellaneous IDA-only methods.
	virtual void AuditTailChunkOwnership(void)
	{ assert(0); return; /* shouldn't be called  */ }
    	virtual void AuditCodeTargets(void)
	{  assert(0); return; /* shouldn't be called.   */ }

	// Patch IDA Pro database.
	virtual bool STARS_patch_byte(STARS_ea_t InstAddr, uint32_t ByteValue) 
	{ assert(0); return 0; /* shouldn't be called */}


	// support for some irdb functions.
	libIRDB::Instruction_t* GetIBT(const libIRDB::AddressID_t& in, bool &success)
	{
		
		std::map<libIRDB::AddressID_t, libIRDB::Instruction_t*>::iterator it=ibta_to_id_map.find(in);
		if(it!=ibta_to_id_map.end())
		{
			success=true;
			return it->second;
		}
		else
		{
			success=false;
			return NULL;
		}
	}

	
private:

	void InitSegments()
	{
		int elfoid=firp->GetFile()->GetELFOID();
		pqxx::largeobject lo(elfoid);
		lo.to_file(pqxx_interface.GetTransaction(),"readeh_tmp_file.exe");
//		elfiop.reset(new ELFIO::elfio);
		elfiop=new ELFIO::elfio;
		elfiop->load("readeh_tmp_file.exe");

		ELFIO::dump::header(std::cout,*elfiop);
		ELFIO::dump::section_headers(std::cout,*elfiop);

		segments.resize(elfiop->sections.size());
		int j=0;
		for(int i=0;i<elfiop->sections.size(); i++)
		{
			if((elfiop->sections[i]->get_flags() & SHF_ALLOC) == SHF_ALLOC)
			{
				std::cout<<"Found 'segment'["<<std::dec<<i<<"] named "<<elfiop->sections[i]->get_name();
				std::cout<<" start: "<<std::hex<<elfiop->sections[i]->get_address()<<std::endl;
				segments[j]=new STARS_IRDB_Segment_t(elfiop->sections[i],j);
				j++;
			}
		}
		segments.resize(j);

	}

	void InitFunctions()
	{
		functions.resize(firp->GetFunctions().size());
		int i=0;

		// create a stars function for each IRDB function
		for(libIRDB::FunctionSet_t::iterator it=firp->GetFunctions().begin();
			it!=firp->GetFunctions().end();
			++it)
		{
			libIRDB::Function_t* f=*it;
			assert(f);
			if(f->GetEntryPoint()==NULL)
				continue;
			STARS_IRDB_Function_t* sf=new STARS_IRDB_Function_t(f, firp);
			assert(sf);

			// also remember a mapping of stars function to function number.
			functions[i++]=sf;

			// and record the entry into a map that let's us look up the function of each instruction id
			for(libIRDB::InstructionSet_t::iterator it=f->GetInstructions().begin();
				it!=f->GetInstructions().end();
				++it)
			{
				libIRDB::Instruction_t* i=*it;
				assert(i);
				instr_id_to_func_map[i->GetBaseID()]=sf;
			}
		}
	}

	void InitIBTAMaps()
	{
		for(libIRDB::InstructionSet_t::iterator it=firp->GetInstructions().begin();
			it!=firp->GetInstructions().end();
			++it)
		{
			libIRDB::Instruction_t* insn=*it;
			assert(insn);
			if(insn->GetIndirectBranchTargetAddress())
			{
				ibta_to_id_map [ *insn->GetIndirectBranchTargetAddress() ] = insn;
			}
		}
		
	}

	std::vector<STARS_IRDB_Segment_t*> segments;
	std::map<STARS_ea_t, STARS_IRDB_Function_t*> instr_id_to_func_map;
	std::map<libIRDB::AddressID_t, libIRDB::Instruction_t*> ibta_to_id_map;
	std::vector<STARS_IRDB_Function_t*> functions;
	ELFIO::elfio* elfiop;


	

	libIRDB::FileIR_t *firp;
	libIRDB::pqxxDB_t &pqxx_interface;

	libIRDB::InstructionPredecessors_t instruction_preds;

	friend class STARS_IRDB_Function_t;
};


#endif


